from utils.fobj_val import fobj_val
import torch
from torch import nn
from utils.Jexc import exc_operator
from utils.Jorthogonal_test import Jtest
import time
from utils.mdot import mdot
import torch.optim as optim

torch.manual_seed(0)
class UMCM_al(nn.Module):
    def __init__(self, X, config_yaml, stepsize):
        super().__init__()
        self.X = nn.Parameter(data=X)
        self.maxiter = int(float(config_yaml["run"]["maxiter"]))
        self.d = config_yaml["datafeature"]["d"]
        self.p = config_yaml["datafeature"]["p"]
        self.stopt = config_yaml["run"]["stop_t"]
        self.stoplr = int(float(config_yaml["run"]["stop_lr"]))
        self.Jerrindex = config_yaml["run"]["Jerrindex"]
        self.optimizer = optim.Adagrad([self.X], lr=float(stepsize))

        J = torch.eye(self.d)
        J[self.p:, self.p:] = -1 * torch.eye(self.d - self.p)
        self.J = J.to(torch.float32)
        self.myeps = 1e-8
        self.beta = 1e4

        self.Jerr0 = Jtest(self.X, self.p)

    def Train(self, C):
        hist_Obj = torch.zeros(self.maxiter, 1)
        hist_t = torch.zeros(self.maxiter, 1)
        hist_err = torch.zeros(self.maxiter, 1)
        start_time = time.time()

        for iter in range(self.maxiter):
            hist_Obj[iter] = fobj_val(self.X, C)
            hist_err[iter] = Jtest(self.X, self.p)

            self.step_forward(C)

            if iter % 2 == 0:
                if self.beta < 1e12:
                    self.beta *= 2

            hist_t[iter] = time.time() - start_time
            if hist_t[iter] > self.stopt:
                break

        hist_Obj = hist_Obj[hist_Obj != 0]
        hist_t = hist_t[:len(hist_Obj)]
        hist_err = hist_err[:len(hist_Obj)]
        return hist_Obj, self.X, hist_t, hist_err

    def step_forward(self, C):
        gradX = C @ self.X
        fval = fobj_val(self.X, C) - 0.5*mdot(self.X.t() @ self.J @ self.X - self.J,
                                                    self.J @ self.X.t() @ gradX) + 0.5*self.beta*torch.norm(self.X.t() @ self.J @ self.X - self.J, 'fro')**2
        fval.backward()
        self.optimizer.step() #进行一次梯度下降
        self.X.grad = None

        self.X.data = torch.clamp(self.X, -100, 100)


