import torch
from torch import nn
import torch.optim as optim
from tqdm import tqdm
from utils.Jexc import exc_operator
from utils.Jorthogonal_test import Jtest
import time

from utils.mdot import mdot
from UHstruct.fobj_val import UltraE_fval_obj
from UHstruct.get_hit_ranks import UltraE_test_hits_rank

torch.manual_seed(0)
class UMCM_al(nn.Module):
    def __init__(self, T, H, Q, P, config_yaml, stepsize):
        super().__init__()
        self.device = config_yaml["device"]
        self.T = T.to(self.device)
        self.H = H.to(self.device)
        self.Q = nn.Parameter(data=Q.to(self.device))
        self.P = P.to(self.device)

        self.config_yaml = config_yaml
        self.maxiter = int(float(config_yaml["run"]["maxiter"]))
        self.Cmaxiter = int(float(config_yaml["run"]["Cmaxiter"]))
        self.d = config_yaml["datafeature"]["d"]
        self.p = config_yaml["datafeature"]["p"]
        self.stopt = config_yaml["run"]["stop_t"]
        self.stoplr = float(config_yaml["run"]["stop_lr"])
        self.stopindex = 0

        self.maxiter = int(float(config_yaml["run"]["maxiter"]))
        self.d = config_yaml["datafeature"]["d"]
        self.p = config_yaml["datafeature"]["p"]
        self.stopt = config_yaml["run"]["stop_t"]

        self.optimizer = optim.Adagrad([self.Q], lr=float(stepsize))

        J = torch.eye(self.d)
        J[self.p:, self.p:] = -1 * torch.eye(self.d - self.p)
        self.J = J.to(torch.float32).to(self.device)
        self.myeps = 1e-8
        self.beta = 1e2 #不收敛要调大这个

    def Train(self):
        hist_Obj = torch.zeros(self.maxiter, 1).to(self.device)
        hist_t = torch.zeros(self.maxiter, 1).to(self.device)
        hist_err = torch.zeros(self.maxiter, 1)

        start_time = time.time()

        for iter in tqdm(range(self.maxiter), desc='Outiter'):
            with torch.no_grad():
                hist_Obj[iter] = UltraE_fval_obj(self.T, self.H, self.P, self.Q, self.config_yaml)
                hist_err[iter] = Jtest(self.Q, self.p)

            self.stepforward()

            if iter % 2 == 0:
                if self.beta < 1e12:
                    self.beta *= 2

            with torch.no_grad():
                hist_t[iter] = time.time() - start_time
                hist_err[iter] = Jtest(self.Q, self.p)
                if hist_t[iter] > self.stopt:
                    self.stopindex = 1

                if self.stopindex == 1:
                    break


        hist_Obj = hist_Obj[hist_Obj != 0]
        hist_t = hist_t[:len(hist_Obj)]
        hist_err = hist_err[:len(hist_Obj)]
        return hist_Obj, self.Q, hist_t, hist_err

    def stepforward(self):
        fval = UltraE_fval_obj(self.T, self.H, self.P, self.Q, self.config_yaml)


        # self.Q.grad = torch.autograd.grad(fval, self.Q, create_graph=True, retain_graph=True)[0]

        fval.backward()

        if torch.isnan(self.Q.grad).any():
            self.Q.grad[torch.isnan(self.Q.grad)] = 0
        if torch.isinf(self.Q.grad).any():
            self.Q.grad[torch.isinf(self.Q.grad)] = 0

        self.updateR()
        self.Q.grad = None

    def updateR(self):
        fval = UltraE_fval_obj(self.T, self.H, self.P, self.Q, self.config_yaml) - 0.5*mdot(self.Q.t() @ self.J @ self.Q - self.J,
                                                    self.J @ self.Q.t() @ self.Q.grad) + 0.5*self.beta*torch.norm(self.Q.t() @ self.J @ self.Q - self.J, 'fro')**2
        fval.backward()
        '''if torch.any(torch.isnan(self.Q.grad)):
            aaa = 1'''

        if torch.isnan(self.Q.grad).any():
            self.Q.grad[torch.isnan(self.Q.grad)] = 0
        if torch.isinf(self.Q.grad).any():
            self.Q.grad[torch.isinf(self.Q.grad)] = 0

        self.optimizer.step() #进行一次梯度下降
        self.Q.grad = None

        self.Q.data = torch.clamp(self.Q, -100, 100)





